@@ -15,8 +15,10 @@ import itertools
 import json
 import re
 import warnings
-
+import enum
 from abc import ABCMeta, abstractmethod
+from datetime import datetime
+from decimal import Decimal
 from typing import Any, Generator, List, Optional, Tuple, Union
 
 import google.protobuf as _
@@ -45,6 +47,9 @@ class BaseMessage(metaclass=ABCMeta):
     def __init__(self, message):
         self.raw = message
 
+    # @abstractmethod
+    # def to_dict(self): ...
+
 class IterableMessage(BaseMessage):
 
     @abstractmethod
@@ -60,118 +65,42 @@ class IterableMessage(BaseMessage):
             collected.append(item)
         return collected
 
-class CapabilitiesResponse_(BaseMessage):
-    r"""Represents a gnmi.CapabilitiesResponse message
+
+class Notification_(IterableMessage):
+    r"""Represents a gnmi.Notification message
 
     """
+    
+    def __iter__(self) -> Generator[Union['Update_', 'Path_'], None, None]:
+        return itertools.chain(self.update, self.delete)
 
     @property
-    def supported_models(self):
-        for model in self.raw.supported_models:
-            yield {
-                "name": model.name,
-                "organization": model.organization,
-                "version": model.version
-            }
-    models = supported_models
+    def atomic(self) -> bool:
+        return self.raw.atomic
 
     @property
-    def supported_encodings(self):
-        return self.raw.supported_encodings
-    encodings = supported_encodings
-
-    @property
-    def gnmi_version(self):
-        return self.raw.gNMI_version
-    version = gnmi_version
-
-@deprecated(("Deprecated in favour of using the "
-    "google.golang.org/genproto/googleapis/rpc/status"
-    "message in the RPC response."))
-class Error_(BaseMessage):
+    def prefix(self) -> 'Path_':
+        return Path_(self.raw.prefix)
     
     @property
-    def code(self) -> int:
-        return self.raw.code
-    
-    @property
-    def message(self) -> str:
-        return self.raw.message
+    def timestamp(self) -> int:
+        return self.raw.timestamp
 
     @property
-    def data(self) -> Any:
-        return self.raw.data
-
-class TypedValue_(BaseMessage):
+    def time(self) -> datetime:
+        return datetime.fromtimestamp(self.timestamp // 1000000000)
+        
+    @property
+    def update(self) -> Generator['Update_', None, None]:
+        for u in self.raw.update:
+            yield Update_(u)
+    updates = update
 
     @property
-    def value(self) -> Any:
-        return self.extract_val()
-
-    def __str__(self):
-        return str(self.extract_val())
-
-    def extract_val(self) -> Any:
-        val = None
-        if self.raw.HasField("any_val"):
-            val = self.raw.any_val
-        elif self.raw.HasField("ascii_val"):
-            val = self.raw.ascii_val
-        elif self.raw.HasField("bool_val"):
-            val = self.raw.bool_val
-        elif self.raw.HasField("bytes_val"):
-            val = self.raw.bytes_val
-        elif self.raw.HasField("decimal_val"):
-            val = self.raw.decimal_val
-        elif self.raw.HasField("float_val"):
-            val = self.raw.float_val
-        elif self.raw.HasField("int_val"):
-            val = self.raw.int_val
-        elif self.raw.HasField("json_ietf_val"):
-            val = json.loads(self.raw.json_ietf_val)
-        elif self.raw.HasField("json_val"):
-            val = json.loads(self.raw.json_val)
-        elif self.raw.HasField("leaflist_val"):
-            lst = []
-            for elem in self.raw.leaflist_val.element:
-                lst.append(TypedValue_(elem).extract_val())
-            val = lst
-        elif self.raw.HasField("proto_bytes"):
-            val = self.raw.proto_bytes
-        elif self.raw.HasField("string_val"):
-            val = self.raw.string_val
-        elif self.raw.HasField("uint_val"):
-            val = self.raw.uint_val
-        else:
-            raise ValueError("Unhandled typed value %s" % self.raw)
-
-        return val
-
-@deprecated("Message 'Value' is deprecated and may be removed in the future")
-class Value_(BaseMessage):
-
-    @property
-    def value(self):
-        return self.extract_val()
-
-    @property
-    def type(self):
-        return self.raw.type
-
-    def __str__(self):
-        return str(self.extract_val())
-
-    def extract_val(self) -> Any:
-        val = None
-        if self.type in (pb.JSON_IETF, pb.JSON) and self.raw.value:
-            val = json.loads(self.raw.value)
-        elif self.type in (pb.BYTES, pb.PROTO):
-            val = self.raw.value
-        elif self.type == pb.ASCII:
-            val = str(self.raw.value)
-        else:
-            raise ValueError("Unhandled type of value %s" % str(val))
-        return val
+    def delete(self) -> Generator['Path_', None, None]:
+        for path in self.raw.delete:
+            yield Path_(path)
+    deletes = delete
 
 class Update_(BaseMessage):
     r"""Represents a gnmi.Update message
@@ -200,12 +129,12 @@ class Update_(BaseMessage):
         return Path_(self.raw.path)
 
     @property
-    def val(self) -> Optional[TypedValue_]:
+    def val(self) -> Optional['TypedValue_']:
         if self.raw.HasField('val'):
             return TypedValue_(self.raw.val)
     
     @property
-    def value(self) -> Optional[Value_]:
+    def value(self) -> Optional['Value_']:
         if self.raw.HasField('value'):
             return Value_(self.raw.value)
     
@@ -213,7 +142,7 @@ class Update_(BaseMessage):
     def duplicates(self):
         return self.raw.duplicates
 
-    def get_value(self) -> Union[TypedValue_, Value_]:
+    def get_value(self) -> Union['TypedValue_', 'Value_']:
         if self.val:
             return self.val.extract_val()
         elif self.value:
@@ -241,140 +170,50 @@ class Update_(BaseMessage):
         
         return cls(pb.Update(path=path.raw, val=typed_value))
 
-
-class Notification_(IterableMessage):
-    r"""Represents a gnmi.Notification message
-
-    """
-    
-    def __iter__(self):
-        return self.update
+class TypedValue_(BaseMessage):
 
     @property
-    def prefix(self):
-        return Path_(self.raw.prefix)
-    
-    @property
-    def timestamp(self):
-        return self.raw.timestamp
+    def value(self) -> Any:
+        return self.extract_val()
 
-    @property
-    def update(self) -> Generator[Update_, None, None]:
-        for u in self.raw.update:
-            yield Update_(u)
-
-    updates = update
-
-class GetResponse_(IterableMessage):
-    r"""Represents a gnmi.GetResponse message
-
-    """
-
-    def __iter__(self):
-        return self.notification
-
-    @property
-    def notification(self):
-        for notification in self.raw.notification:
-            yield Notification_(notification)
-    notifications = notification
-        
-    @property
-    def error(self) -> Error_:
-        if self.raw.HasField('error'):
-            return Error_(self.raw.error)
-
-class UpdateResult_(BaseMessage):
-    _OPERATION = [
-        "INVALID",
-        "DELETE",
-        "REPLACE",
-        "UPDATE"
-    ]
-    
     def __str__(self):
-        return "%s %s" % (self.operation, self.path)
-    
-    @property
-    def op(self):
-        #return self._OPERATION.index(self.raw.operation) 
-        
-        return self._OPERATION[self.raw.op]
-    operation = op
-    
-    @property
-    def path(self):
-        return Path_(self.raw.path)
+        return str(self.extract_val())
 
-    @property
-    @deprecated(("Deprecated timestamp for the UpdateResult, this field has been "
-        "replaced by the timestamp within the SetResponse message, since "
-        "all mutations effected by a set should be applied as a single "
-        "transaction."))
-    def timestamp(self) -> int:
-        return self.raw.timestamp
+    def extract_val(self) -> Any:
+        val = None
+        if self.raw.HasField("any_val"):
+            val = self.raw.any_val
+        elif self.raw.HasField("ascii_val"):
+            val = self.raw.ascii_val
+        elif self.raw.HasField("bool_val"):
+            val = self.raw.bool_val
+        elif self.raw.HasField("bytes_val"):
+            val = self.raw.bytes_val
+        elif self.raw.HasField("decimal_val"):
+            val = self.raw.decimal_val
+            val = Decimal(str(val.digits / 10**val.precision))
+        elif self.raw.HasField("float_val"):
+            val = self.raw.float_val
+        elif self.raw.HasField("int_val"):
+            val = self.raw.int_val
+        elif self.raw.HasField("json_ietf_val"):
+            val = json.loads(self.raw.json_ietf_val)
+        elif self.raw.HasField("json_val"):
+            val = json.loads(self.raw.json_val)
+        elif self.raw.HasField("leaflist_val"):
+            val = []
+            for elem in self.raw.leaflist_val.element:
+                val.append(TypedValue_(elem).extract_val())
+        elif self.raw.HasField("proto_bytes"):
+            val = self.raw.proto_bytes
+        elif self.raw.HasField("string_val"):
+            val = self.raw.string_val
+        elif self.raw.HasField("uint_val"):
+            val = self.raw.uint_val
+        else:
+            raise ValueError("Unhandled typed value %s" % self.raw)
 
-    @property
-    def error(self) -> Optional[Error_]:
-        if self.raw.HasField('error'):
-            return Error_(self.raw.error)
-
-class SetResponse_(IterableMessage):
-
-    def __iter__(self):
-        return self.response
-
-    @property
-    def timetamp(self):
-        pass
-
-    @property
-    def response(self):
-        for resp in self.raw.response:
-            yield UpdateResult_(resp)
-
-    responses = response
-
-    @property
-    def error(self) -> Optional[Error_]:
-        if self.raw.HasField('error'):
-            return Error_(self.raw.error)
-
-class SubscribeResponse_(BaseMessage):
-    r"""Represents a gnmi.SubscribeResponse message
-
-    """
-
-    # @property
-    # def sync_response(self):
-    #     pass
-    @property
-    def sync_response(self) -> bool:
-        return self.raw.sync_response
-
-    @property
-    def update(self):
-        return Notification_(self.raw.update)
-
-    @property
-    def error(self) -> Optional[Error_]:
-        if self.raw.HasField('error'):
-            return Error_(self.raw.error)
-
-class PathElem_(BaseMessage):
-    r"""Represents a gnmi.PathElem message
-
-    """
-    
-    @property
-    def key(self):
-        if hasattr(self.raw, "key"):
-            return self.raw.key 
-        return {}
-
-    @property
-    def name(self):
-        return self.raw.name
+        return val
 
 class Path_(IterableMessage):
     r"""Represents a gnmi.Path message
@@ -403,7 +242,7 @@ class Path_(IterableMessage):
         return self.elem
 
     @property
-    def elem(self) -> Generator[PathElem_, None, None]:
+    def elem(self) -> Generator['PathElem_', None, None]:
         for elem in self.raw.elem:
             yield PathElem_(elem)
     elems = elem
@@ -433,8 +272,8 @@ class Path_(IterableMessage):
                 val = util.escape_string(val, "]")
                 path += "[" + key + "=" + val + "]"
 
-        if not path:
-            path =  "/" + "/".join(self.element)
+        if not path and self.element:
+            path =  "/".join(self.element)
 
         if self.origin:
             path = ":".join([self.origin, path])
@@ -461,9 +300,197 @@ class Path_(IterableMessage):
 
         return cls(pb.Path(origin=origin, elem=elems)) # type: ignore
 
+class PathElem_(BaseMessage):
+    r"""Represents a gnmi.PathElem message
+
+    """
+    
+    @property
+    def key(self):
+        if hasattr(self.raw, "key"):
+            return self.raw.key 
+        return {}
+
+    @property
+    def name(self):
+        return self.raw.name
+
+@deprecated("Message 'Value' is deprecated and may be removed in the future")
+class Value_(BaseMessage):
+
+    @property
+    def value(self):
+        return self.extract_val()
+
+    @property
+    def type(self):
+        return self.raw.type
+
+    def __str__(self):
+        return str(self.extract_val())
+
+    def extract_val(self) -> Any:
+        val = None
+        if self.type in (pb.JSON_IETF, pb.JSON) and self.raw.value:
+            val = json.loads(self.raw.value)
+        elif self.type in (pb.BYTES, pb.PROTO):
+            val = self.raw.value
+        elif self.type == pb.ASCII:
+            val = str(self.raw.value)
+        else:
+            raise ValueError("Unhandled type of value %s" % str(val))
+        return val
+
+class Encoding_(enum.Enum):
+    JSON = 0
+    BYTES = 1
+    PROTO = 2
+    ASCII = 3
+    JSON_IETF = 4
+
+@deprecated(("Deprecated in favour of using the "
+    "google.golang.org/genproto/googleapis/rpc/status"
+    "message in the RPC response."))
+class Error_(BaseMessage):
+    
+    @property
+    def code(self) -> int:
+        return self.raw.code
+    
+    @property
+    def message(self) -> str:
+        return self.raw.message
+
+    @property
+    def data(self) -> Any:
+        return self.raw.data
+
+class SubscribeResponse_(BaseMessage):
+    r"""Represents a gnmi.SubscribeResponse message
+
+    """
+    
+    @property
+    def sync_response(self) -> bool:
+        return self.raw.sync_response
+
+    @property
+    def update(self):
+        return Notification_(self.raw.update)
+    notification = update
+
+    @property
+    def error(self) -> Optional[Error_]:
+        if self.raw.HasField('error'):
+            return Error_(self.raw.error)
+
+class SetResponse_(IterableMessage):
+
+    def __iter__(self):
+        return self.response
+
+    @property
+    def timetamp(self):
+        pass
+
+    @property
+    def response(self):
+        for resp in self.raw.response:
+            yield UpdateResult_(resp)
+
+    responses = response
+
+    @property
+    def error(self) -> Optional[Error_]:
+        if self.raw.HasField('error'):
+            return Error_(self.raw.error)
+
+# class Operation_(enum.Enum):
+#     "INVALID" = 0
+#     "DELETE" = 1
+#     "REPLACE" = 2
+#     "UPDATE" = 3
+
+class UpdateResult_(BaseMessage):
+    _OPERATION = [
+        "INVALID",
+        "DELETE",
+        "REPLACE",
+        "UPDATE"
+    ]
+    
+    def __str__(self):
+        return "%s %s" % (self.operation, self.path)
+    
+    @property
+    def op(self):
+        return self._OPERATION[self.raw.op]
+    operation = op
+    
+    @property
+    def path(self):
+        return Path_(self.raw.path)
+
+    @property
+    @deprecated(("Deprecated timestamp for the UpdateResult, this field has been "
+        "replaced by the timestamp within the SetResponse message, since "
+        "all mutations effected by a set should be applied as a single "
+        "transaction."))
+    def timestamp(self) -> int:
+        return self.raw.timestamp
+
+    @property
+    def error(self) -> Optional[Error_]:
+        if self.raw.HasField('error'):
+            return Error_(self.raw.error)
+
+class GetResponse_(IterableMessage):
+    r"""Represents a gnmi.GetResponse message
+
+    """
+
+    def __iter__(self):
+        return self.notification
+
+    @property
+    def notification(self):
+        for notification in self.raw.notification:
+            yield Notification_(notification)
+    notifications = notification
+        
+    @property
+    def error(self) -> Optional[Error_]:
+        if self.raw.HasField('error'):
+            return Error_(self.raw.error)
+
+class CapabilitiesResponse_(BaseMessage):
+    r"""Represents a gnmi.CapabilitiesResponse message
+
+    """
+
+    @property
+    def supported_models(self):
+        for model in self.raw.supported_models:
+            yield {
+                "name": model.name,
+                "organization": model.organization,
+                "version": model.version
+            }
+    models = supported_models
+
+    @property
+    def supported_encodings(self):
+        return [Encoding_(i) for i in self.raw.supported_encodings]
+    encodings = supported_encodings
+
+    @property
+    def gnmi_version(self):
+        return self.raw.gNMI_version
+    version = gnmi_version
+
 class Status_(collections.namedtuple('Status_', 
         ('code', 'details', 'trailing_metadata')), grpc.Status):
     
     @classmethod
     def from_call(cls, call):
-        return cls(call.code(), call.details(), call.trailing_metadata())
+        return cls(call.code(), call.details(), call.trailing_metadata())
\ No newline at end of file
